--Random SPB
--by Callmore#4947

local cv_defaultspbenabled = CV_RegisterVar{
	name = "rspb_defaultspbenabled",
	defaultvalue = "Yes",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
}
local cv_spbforce = CV_RegisterVar{
	name = "rspb_spbforce",
	defaultvalue = 0,
	flags = CV_NETVAR,
	PossibleValue = CV_Unsigned,
}
local cv_fade = CV_RegisterVar{
	name = "rspb_fade",
	defaultvalue = "Always",
	flags = CV_NETVAR,
	PossibleValue = {Never=0, Tall=1, Always=2},
}
local cv_enable = CV_RegisterVar{
	name = "rspb_enable",
	defaultvalue = "Yes",
	flags = 0,
	PossibleValue = CV_YesNo,
}

local SPB_HIDE_RANGE = 128*FRACUNIT
local VANILLA_SPB = 0

rawset(_G, "rspb", {
	replacementtable = {
		[0] = {
			sprite = SPR_SPBM,
			dead = 8,
			tall = false,
			frames = {0, 1, 0, 2, 0, 3, 0, 4, 0, 5, 0, 6, 0, 7, 0, 8, 0, 8, 0, 8},
			sound = sfx_kc57,
		},
	},
})

-- Code for allowing the user to permemently remove an spb from the rotation
-- the sound will still stay behind but thats a single sound freeslot
local function removeSPB(p, id)
	id = tonumber(id)
	if id == nil then
		CONS_Printf(p, "Remove an SPB from the currently loaded permemently. All IDs after this ID will be shifted down by 1")
		return
	end
	if id < 0 then
		CONS_Printf(p, "SPBs cannot have IDs under 0.")
	elseif id == 0 then
		CONS_Printf(p, 'Cannot delete the default SPB, set \130rspb_defaultspbenabled\128 to "No" instead.')
	elseif id > #rspb.replacementtable then
		CONS_Printf(p, "ID is over the amount of remaining SPBs.")
	else
		table.remove(rspb.replacementtable, id)
	end
end
COM_AddCommand("rspb_removespb", removeSPB, true)

-- Replacing the entire state system with a logical less statey system
freeslot("S_RSPB_SPB", "S_RSPB_DEAD")

states[S_RSPB_SPB] = {SPR_THOK, A, -1, nil, 0, 0, S_RSPB_SPB}
states[S_RSPB_DEAD] = {SPR_THOK, A, 175, nil, 0, 0, S_NULL}

local spr_attempts = {
	"SPR_(.?.?.?.?)",
	"(.?.?.?.?)",
}

local function extractSprFreeslot(s)
	s = s:upper()

	for _, p in ipairs(spr_attempts) do
		if s:match(p) then
			return "SPR_" .. s:match(p)
		end
	end
end

local sfx_attempts = {
	"sfx_(.?.?.?.?.?.?)",
	"ds(......)",
	"(.?.?.?.?.?.?)",
}

local function extractSfxFreeslot(s)
	s = s:lower()
	
	for _, p in ipairs(sfx_attempts) do
		if s:match(p) then
			return "sfx_" .. s:match(p)
		end
	end
end

function rspb.addReplacement(spriteName, animation, deadFrame, soundName, isTall)
	-- convert the id into useable strings
	local sprite
	if type(spriteName) == "string" then
		local str_sprite = extractSprFreeslot(spriteName)
		if str_sprite == nil then
			print("ERROR: Failed to parse \"" .. spriteName .. "\" into valid freeslot name.")
			return
		end
		-- make freeslots
		freeslot(str_sprite)
		-- get freeslots into variables
		sprite = _G[str_sprite]
	elseif type(spriteName) == "number" then
		sprite = spriteName
	else
		print("ERROR: Invalid sprite name \"" .. tostring(spriteName) .. "\"")
		return
	end
	
	-- construct replacement
	local replacement_table = {
		sprite = sprite,
		dead = deadFrame or 1,
		tall = isTall or false,
		frames = animation,
	}
	
	-- scream
	local str_sfxScream
	if type(soundName) == "string" then
		str_sfxScream = extractSfxFreeslot(soundName)
		if str_sfxScream == nil then
			print("ERROR: Failed to parse \"" .. soundName .. "\" into valid freeslot name.")
			return
		end
	elseif type(soundName) == "number" then
		-- assume they already freeslotted it
		sfxinfo[soundName].singular = false
		sfxinfo[soundName].priority = 128
		sfxinfo[soundName].flags = 8

		replacement_table.sound = soundName
	end
	
	if str_sfxScream ~= nil then
		freeslot(str_sfxScream)
		local sfxScream = _G[str_sfxScream]
		sfxinfo[sfxScream].singular = false
		sfxinfo[sfxScream].priority = 128
		sfxinfo[sfxScream].flags = 8

		replacement_table.sound = sfxScream
	end

	rspb.replacementtable[#rspb.replacementtable+1] = replacement_table
end

local function between(x, mi, ma)
	return x >= mi and x <= ma
end

-- SPB
local function spbSpawnRandom(mo)
	if not cv_spbforce.value then
		local generated = nil
		if cv_defaultspbenabled.value then
			generated = P_RandomKey(#rspb.replacementtable+1)
		else
			generated = P_RandomKey(#rspb.replacementtable)+1
		end
		if cv_enable.value then
			mo.rspb_rnum = generated
		else
			mo.rspb_rnum = VANILLA_SPB
		end
	else --we forced picked an spb
		if cv_spbforce.value > VANILLA_SPB and cv_spbforce.value <= #rspb.replacementtable then
			mo.rspb_rnum = cv_spbforce.value
		else
			print(cv_spbforce.value .. " is not a valid SPB number!")
			mo.rspb_rnum = VANILLA_SPB
		end
	end
	
	-- Initialise the timer
	mo.rspb_timer = 0
end
addHook("MobjSpawn", spbSpawnRandom, MT_SPB)

local function spbSwapper(mo)
	if mo.rspb_rnum > #rspb.replacementtable then
		mo.rspb_rnum = VANILLA_SPB
		print("SPB replacement went missing! Reverting SPB back to vanilla")
	end
	
	if not cv_enable.value and mo.rspb_rnum ~= VANILLA_SPB then
		mo.rspb_rnum = VANILLA_SPB
	end

	local spbinfo = rspb.replacementtable[mo.rspb_rnum]
	if spbinfo == nil then
		-- panic
		mo.rspb_rnum = 0
		spbinfo = rspb.replacementtable[mo.rspb_rnum]
	end
	
	if not mo.rspb_default and between(mo.state, S_SPB1, S_SPB20) then
		mo.state = S_RSPB_SPB
		mo.rspb_default = true
	elseif not mo.rspb_dead and mo.state == S_SPB_DEAD then
		mo.state = S_RSPB_DEAD
		mo.sprite = spbinfo.sprite
		mo.frame = spbinfo.dead
		mo.rspb_dead = true
	end
	
	if mo.state == S_RSPB_SPB then
		A_SPBChase(mo, 0, 0)
		mo.sprite = spbinfo.sprite
		mo.frame = spbinfo.frames[(mo.rspb_timer % #spbinfo.frames) + 1]
		mo.rspb_timer = $ + 1
	end
end
addHook("MobjThinker", spbSwapper, MT_SPB)

local function spbSound(mo)
	if mo.rspb_rnum ~= VANILLA_SPB
	and rspb.replacementtable[mo.rspb_rnum].sound
	and S_SoundPlaying(mo, sfx_kc57)
	and cv_enable.value then
		S_StopSoundByID(mo, sfx_kc57)
		S_StartSound(mo, rspb.replacementtable[mo.rspb_rnum].sound)
	end
end
addHook("MobjThinker", spbSound, MT_SPB)

local function spbTransparency(mo)
	if mo and mo.valid then
		if ((cv_fade.value == 2) or (rspb.replacementtable[mo.rspb_rnum].tall and cv_fade.value ~= 0))
		and R_PointToDist(mo.x, mo.y) < 128*FRACUNIT then
			mo.flags2 = $ | MF2_SHADOW
		else
			mo.flags2 = $ & ~MF2_SHADOW
		end
	end
end
addHook("MobjThinker", spbTransparency, MT_SPB)
